
##
## Environment utilities
##

.BaseEnv <- if (exists("baseenv")) baseenv() else NULL
.EmptyEnv <- if (exists("emptyenv")) emptyenv() else NULL

is.emptyenv <- function(e) identical(e, .EmptyEnv)
is.baseenv <- function(e) identical(e, .BaseEnv)

mkHash <- function() new.env(hash = TRUE, parent = .EmptyEnv)

getAssignedVar <- function(e) {
    v <- e[[2]];
    while (typeof(v) == "language")
        v <- v[[2]]
    if (typeof(v) != "symbol" && typeof(v) != "character")
        stop(paste("bad assignment:", as.expression(e)))
    as.character(v)
}
findLocals <- function(e) {
    if (typeof(e) == "language") {
        if (typeof(e[[1]]) == "symbol")
            switch(as.character(e[[1]]),
                   "=" =,
                   "<-" = unique(c(getAssignedVar(e),
                                   findLocals(e[[2]]), findLocals(e[[3]]))),
                    "local" = NULL, #**** only if no env arg
                    "expression" = NULL, #**** check env
                    "function" = NULL,   
                    "quote" = NULL,      #**** check env
                    "~" = NULL,
                    "for" = unique(c(as.character(e[2]),
                                     unlist(lapply(e[-2], findLocals)))),
                    unique(unlist(lapply(e[-1], findLocals))))
         else unique(unlist(lapply(e, findLocals)))
    }
}
funEnv <- function(forms, body, env) {
    list(unique(c(names(forms), findLocals(body), env[[1]])), env[[2]])
}

matchName <- function(name, list) match(as.character(name), list, 0)

findVar <- function(e, env) matchName(e, env[[1]])

isSysVar <- function(e, env) {
    n <- as.character(e)
    if (findVar(e, env) || ! exists(n, .GlobalEnv))
        FALSE
    else {
        env <- .GlobalEnv
        while (! is.baseenv(env)) {
            if (exists(n, env, inherits = FALSE)) {
                pn <- attr(env, "name")
                if (! is.null(pn) && 
                    (pn == "package:stats" || pn == "package:utils" ||
                     pn == "package:methods"))
                    return(TRUE)
                else return(FALSE)
            }
            else env <- parent.env(env)
        }
        TRUE
    }
}        
foldFuns <- c("+", "-", "*", "/", "^", "(",
              ">", ">=", "==", "!=", "<", "<=", "||", "&&", "!",
              "|", "&", "%%",
              "sqrt", "log", 
              "c", "as.integer", "vector", "integer","numeric","character",
              "rep",
              ":")

constNames <- c("pi", "T", "F")

maxConstSize <- 10

checkConst <- function(e) {
    if ((mode(e) == "numeric" || mode(e) == "logical")
         && length(e) <= maxConstSize)
        e
    else
        NULL
}

constantFold <- function(e, env = NULL) {
    if (typeof(e) == "language") {
        if (length(e) > 1) {
            fun <- e[[1]]
            if (typeof(fun) == "symbol" &&
                matchName(fun, foldFuns) && isSysVar(fun, env)) {
                args <- as.list(e[-1])
                for (i in 1:length(args)) {
                    val <- constantFold(args[[i]], env)
                    if (! is.null(val)) args[[i]] <- val
                    else return(NULL)
                }
                types <- unlist(lapply(args, mode))
                if (all(types == "numeric" | types == "logical")) {
                    # cat(paste("folding", as.expression(e), "\n"));
                    checkConst(do.call(as.character(fun), args))
                }
                else NULL
            }
            else NULL
        }
        else NULL
    }
    else if (typeof(e) == "symbol") {
        if (matchName(e, constNames) && isSysVar(e, env)) {
            # cat(paste("folding", as.expression(e), "\n"));
            checkConst(get(as.character(e)))
        }
        else NULL
    }
    else e
}
Opcodes.argc <- list(
BCMISMATCH.OP= 0,
RETURN.OP= 0,
GOTO.OP= 1,
BRIFNOT.OP= 1,
POP.OP= 0,
DUP.OP= 0,
PRINTVALUE.OP= 0,
STARTLOOPCNTXT.OP= 1,
ENDLOOPCNTXT.OP= 0,
DOLOOPNEXT.OP= 0,
DOLOOPBREAK.OP= 0,
STARTFOR.OP= 2,
STEPFOR.OP= 1,
ENDFOR.OP= 0,
SETLOOPVAL.OP= 0,
INVISIBLE.OP=0,
LDCONST.OP= 1,
LDNULL.OP= 0,
LDTRUE.OP= 0,
LDFALSE.OP= 0,
GETVAR.OP= 1,
DDVAL.OP= 1,
SETVAR.OP= 1,
GETFUN.OP= 1,
GETGLOBFUN.OP= 1,
GETSYMFUN.OP= 1,
GETBUILTIN.OP= 1,
GETINTLBUILTIN.OP= 1,
CHECKFUN.OP= 0,
MAKEPROM.OP= 1,
DOMISSING.OP= 0,
SETTAG.OP= 1,
DODOTS.OP= 0,
PUSHARG.OP= 0,
PUSHCONSTARG.OP= 1,
PUSHNULLARG.OP= 0,
PUSHTRUEARG.OP= 0,
PUSHFALSEARG.OP= 0,
CALL.OP= 1,
CALLBUILTIN.OP= 1,
CALLSPECIAL.OP= 1,
MAKECLOSURE.OP= 1,
UMINUS.OP= 0,
UPLUS.OP= 0,
ADD.OP= 0,
SUB.OP= 0,
MUL.OP= 0,
DIV.OP= 0,
EXPT.OP= 0,
SQRT.OP= 0,
EXP.OP= 0,
EQ.OP= 0,
NE.OP= 0,
LT.OP= 0,
LE.OP= 0,
GE.OP= 0,
GT.OP= 0,
AND.OP= 0,
OR.OP= 0,
NOT.OP= 0,
DOTSERR.OP= 0,
STARTASSIGN.OP= 2,
ENDASSIGN.OP= 2,
STARTSUBSET.OP= 2,
DFLTSUBSET.OP= 0,
STARTSUBASSIGN.OP= 2,
DFLTSUBASSIGN.OP= 0,
STARTC.OP= 2,
DFLTC.OP= 0,
STARTSUBSET2.OP= 2,
DFLTSUBSET2.OP= 0,
STARTSUBASSIGN2.OP= 2,
DFLTSUBASSIGN2.OP= 0,
DOLLAR.OP= 2,
DOLLARGETS.OP= 2,
ISNULL.OP= 0,
ISLOGICAL.OP= 0,
ISINTEGER.OP= 0,
ISDOUBLE.OP= 0,
ISCOMPLEX.OP= 0,
ISCHARACTER.OP= 0,
ISSYMBOL.OP= 0,
ISOBJECT.OP= 0,
ISNUMERIC.OP= 0,
NVECELT.OP= 2,
NMATELT.OP= 3,
SETNVECELT.OP= 3,
SETNMATELT.OP= 4
)

Opcodes.names <- names(Opcodes.argc)

BCMISMATCH.OP <- 0
RETURN.OP <- 1
GOTO.OP <- 2
BRIFNOT.OP <- 3
POP.OP <- 4
DUP.OP <- 5
PRINTVALUE.OP <- 6
STARTLOOPCNTXT.OP <- 7
ENDLOOPCNTXT.OP <- 8
DOLOOPNEXT.OP <- 9
DOLOOPBREAK.OP <- 10
STARTFOR.OP <- 11
STEPFOR.OP <- 12
ENDFOR.OP <- 13
SETLOOPVAL.OP <- 14
INVISIBLE.OP <- 15
LDCONST.OP <- 16
LDNULL.OP <- 17
LDTRUE.OP <- 18
LDFALSE.OP <- 19
GETVAR.OP <- 20
DDVAL.OP <- 21
SETVAR.OP <- 22
GETFUN.OP <- 23
GETGLOBFUN.OP <- 24
GETSYMFUN.OP <- 25
GETBUILTIN.OP <- 26
GETINTLBUILTIN.OP <- 27
CHECKFUN.OP <- 28
MAKEPROM.OP <- 29
DOMISSING.OP <- 30
SETTAG.OP <- 31
DODOTS.OP <- 32
PUSHARG.OP <- 33
PUSHCONSTARG.OP <- 34
PUSHNULLARG.OP <- 35
PUSHTRUEARG.OP <- 36
PUSHFALSEARG.OP <- 37
CALL.OP <- 38
CALLBUILTIN.OP <- 39
CALLSPECIAL.OP <- 40
MAKECLOSURE.OP <- 41
UMINUS.OP <- 42
UPLUS.OP <- 43
ADD.OP <- 44
SUB.OP <- 45
MUL.OP <- 46
DIV.OP <- 47
EXPT.OP <- 48
SQRT.OP <- 49
EXP.OP <- 50
EQ.OP <- 51
NE.OP <- 52
LT.OP <- 53
LE.OP <- 54
GE.OP <- 55
GT.OP <- 56
AND.OP <- 57
OR.OP <- 58
NOT.OP <- 59
DOTSERR.OP <- 60
STARTASSIGN.OP <- 61
ENDASSIGN.OP <- 62
STARTSUBSET.OP <- 63
DFLTSUBSET.OP <- 64
STARTSUBASSIGN.OP <- 65
DFLTSUBASSIGN.OP <- 66
STARTC.OP <- 67
DFLTC.OP <- 68
STARTSUBSET2.OP <- 69
DFLTSUBSET2.OP <- 70
STARTSUBASSIGN2.OP <- 71
DFLTSUBASSIGN2.OP <- 72
DOLLAR.OP <- 73
DOLLARGETS.OP <- 74
ISNULL.OP <- 75
ISLOGICAL.OP <- 76
ISINTEGER.OP <- 77
ISDOUBLE.OP <- 78
ISCOMPLEX.OP <- 79
ISCHARACTER.OP <- 80
ISSYMBOL.OP <- 81
ISOBJECT.OP <- 82
ISNUMERIC.OP <- 83
NVECELT.OP <- 84
NMATELT.OP <- 85
SETNVECELT.OP <- 86
SETNMATELT.OP <- 87
make.codeBuf <- function() {
    code <- list(.Internal(bcVersion()))
    const <- vector("list")
    labels <- vector("list")
    globvars <- vector("list")
    locfuns <- vector("list")
    undefs <- vector("list")
    idx <- 0
    list(code = function() {
             n <- length(code)
             if (n > 0) {
                 for (i in 1:n) {
                     v <- code[[i]]
                     if (is.character(code[[i]]) && ! is.null(labels[[v]]))
                         code[[i]] <- labels[[v]]
                 }
             }
             as.integer(code)
         },
         const = function() const,
         makelabel = function() { idx <<- idx + 1; paste("L", idx, sep="") },
         putcode = function(...) code <<- c(code, list(...)),
         putconst = function(x) {
             n <- length(const)
             if (n > 0) {
                 if (is.symbol(x))
                     for (i in 1:n)
                         if (is.symbol(const[[i]]) && x == const[[i]])
                             return(i-1)
             }
             # if (is.null(x)) const <<- c(const, list(NULL))
             # else const[[n + 1]] <<- x
             const <<- .Internal(putconst(const, x))
             n
         },
         putlabel = function(name) labels[[name]] <<- length(code),
         globvars = function() globvars,
         add.globvar = function(v) if (matchName(v, globvars) == 0)
                           globvars <<- c(globvars, v),
         locfuns = function() locfuns,
         add.locfun = function(v) if (matchName(v, locfuns) == 0)
                          locfuns <<- c(locfuns, v),
         undefs = function() undefs,
         add.undef = function(v) if (matchName(v, undefs) == 0)
                         undefs <<- c(undefs, v),
         warn = function(x) cat(paste("Note:", x, "\n")))
}
dots.or.missing <- function(args) {
    for (i in 1:length(args)) {
        a <-args[[i]]
        if (missing(a)) return(TRUE) #**** better test?
        if (typeof(a) == "symbol" && a == "...") return(TRUE)
    }
    return(FALSE)
}

any.dots <- function(args) {
    for (i in 1:length(args)) {
        a <-args[[i]]
        if (! missing(a) && typeof(a) == "symbol" && a == "...")
            return(TRUE)
    }
    return(FALSE)
}

make.argState <- function(state) {
    list(tailcall=FALSE)
}

make.noValueState <- function(state) {
    s <- state;
    s$tailcall <- FALSE
    s
}

cmpPrim1 <- function(e, cb, env, op, state) {
    if (dots.or.missing(e[-1])) return(FALSE);
    if (length(e) != 2) {
        cb$warn(paste("wrong argument count in ", as.expression(e)))
        return(FALSE)
    }
    s <- make.argState(state)
    cmp(e[[2]], cb, env, s);
    cb$putcode(op)
    if (state$tailcall) cb$putcode(RETURN.OP)
    TRUE
}

cmpPrim2 <- function(e, cb, env, op, state) {
    if (dots.or.missing(e[-1])) return(FALSE);
    if (length(e) != 3) {
        cb$warn(paste("wrong argument count in ", as.expression(e)))
     return(FALSE)
   }
    s <- make.argState(state)
    cmp(e[[2]], cb, env, s);
    cmp(e[[3]], cb, env, s);
    cb$putcode(op)
    if (state$tailcall) cb$putcode(RETURN.OP)
    TRUE
}

primOps <- NULL

primOps$"+" <- function(e, cb, env, state) {
    if (length(e) == 3)
        cmpPrim2(e, cb, env, ADD.OP, state)
    else
        cmpPrim1(e, cb, env, UPLUS.OP, state)
}

primOps$"-" <- function(e, cb, env, state) {
    if (length(e) == 3)
        cmpPrim2(e, cb, env, SUB.OP, state)
    else
        cmpPrim1(e, cb, env, UMINUS.OP, state)
}

primOps$"*" <- function(e, cb, env, state) {
    cmpPrim2(e, cb, env, MUL.OP, state);
}

primOps$"/" <- function(e, cb, env, state) {
    cmpPrim2(e, cb, env, DIV.OP, state);
}

primOps$"^" <- function(e, cb, env, state) {
    cmpPrim2(e, cb, env, EXPT.OP, state);
}

primOps$"(" <- function(e, cb, env, state) {
    if (length(e) != 2) {
        cb$warn("wrong number of arguments in \"(\"")
        return(FALSE)
    }
    cmp(e[[2]], cb, env, state)
    TRUE
}

primOps$"exp" <- function(e, cb, env, state) {
    cmpPrim1(e, cb, env, EXP.OP, state)
}

primOps$"sqrt" <- function(e, cb, env, state) {
    cmpPrim1(e, cb, env, SQRT.OP, state)
}

primOps$"{" <- function(e, cb, env, state) {
    n <- length(e)
    if (n == 1) cmp(NULL, cb, env, state)
    else {
        if (n > 2) {
            s <- make.noValueState(state)
            for (i in 2:(n-1)) {
                cmp(e[[i]], cb, env, s)
                cb$putcode(POP.OP)
            }
        }
        cmp(e[[n]], cb, env, state)
    }
    TRUE
}

#**** check test value for length, NA
primOps$"if" <- function(e, cb, env, state) {
    then.expr <- e[[3]]
    else.expr <- if (length(e) == 4) e[[4]] else quote(invisible(NULL))
    s <- make.argState(state)
    cmp(e[[2]], cb, env, s)
    else.label <- cb$makelabel()
    cb$putcode(BRIFNOT.OP, else.label)
    if (state$tailcall) {
        cmp(then.expr, cb, env, state)
        cb$putlabel(else.label)
        cmp(else.expr, cb, env, state)
    }
    else {
        end.label <- cb$makelabel()
        cmp(then.expr, cb, env, state)
        cb$putcode(GOTO.OP, end.label)
        cb$putlabel(else.label)
        cmp(else.expr, cb, env, state)
        cb$putlabel(end.label)
    }
    TRUE
}
make.loopState <- function(state, loop.label, end.label) {
    s <- make.noValueState(state)
    s$loop <- list(loop=loop.label, end=end.label)
    s
}

primOps$"repeat" <- function(e, cb, env, state) {
    loop.label <- cb$makelabel()
    end.label <- cb$makelabel()
    s <- make.loopState(state, loop.label, end.label)
    cb$putcode(LDNULL.OP)
    cb$putlabel(loop.label)
    cmp(e[[2]], cb, env, s)
    cb$putcode(SETLOOPVAL.OP)
    cb$putcode(GOTO.OP, loop.label)
    cb$putlabel(end.label)
    if (state$tailcall) {
      cb$putcode(INVISIBLE.OP)
      cb$putcode(RETURN.OP)
    }
    TRUE
}

primOps$"break" <- function(e, cb, env, state) {
    if (is.null(state$loop)) {
        cb$warn("break may be used in wrong context: no loop is visible")
        FALSE
    }
    else {
        cb$putcode(GOTO.OP, state$loop$end)
        TRUE
    }
}    

primOps$"next" <- function(e, cb, env, state) {
    if (is.null(state$loop)) {
        cb$warn("next may be used in wrong context: no loop is visible")
        FALSE
    }
    else {
        cb$putcode(GOTO.OP, state$loop$loop)
        TRUE
    }
}    
primOps$"while" <- function(e, cb, env, state) {
    loop.label <- cb$makelabel()
    end.label <- cb$makelabel()
    s <- make.loopState(state, loop.label, end.label)
    cb$putcode(LDNULL.OP)
    cb$putlabel(loop.label)
    cmp(e[[2]], cb, env, s) #**** use argState here??
    cb$putcode(BRIFNOT.OP, end.label)
    cmp(e[[3]], cb, env, s)
    cb$putcode(SETLOOPVAL.OP)
    cb$putcode(GOTO.OP, loop.label)
    cb$putlabel(end.label)
    if (state$tailcall) {
      cb$putcode(INVISIBLE.OP)
      cb$putcode(RETURN.OP)
    }
    TRUE
}
primOps$"for" <- function(e, cb, env, state) {
    body.label <- cb$makelabel()
    loop.label <- cb$makelabel()
    end.label <- cb$makelabel()
    sym <- e[[2]]
    seq <- e[[3]]
    body <- e[[4]]
    if (! is.name(sym)) {
        cb$warn(paste("not a valid for loop argument:", sym));
        FALSE
    }
    s <- make.argState(state)
    cmp(seq, cb, env, s)
    ci <- cb$putconst(sym)
    cb$putcode(STARTFOR.OP, ci, loop.label)
    cb$putlabel(body.label)
    s <- make.loopState(state, loop.label, end.label)
    cmp(body, cb, env, s)
    cb$putcode(SETLOOPVAL.OP)
    cb$putlabel(loop.label)
    cb$putcode(STEPFOR.OP, body.label)
    cb$putlabel(end.label)
     cb$putcode(ENDFOR.OP)
    if (state$tailcall) {
      cb$putcode(INVISIBLE.OP)
      cb$putcode(RETURN.OP)
    }
    TRUE
}
cmpBuiltinArgs <- function(args, names, cb, env, state, keep.missing)
{
    argc <- length(args)
    if (argc > 0) {
        s <- make.argState(state)
        for (i in 1:length(args)) {
            a <- args[[i]]
            n <- names[[i]]
            if (missing(a)) {
                if (keep.missing) cb$putcode(DOMISSING.OP)
                else n <- NULL; #*** skips tag check
            }
            else {
                ca <- constantFold(a, env)
                if (! is.null(ca)) a <- ca
                if (is.symbol(a) || typeof(a) == "language") {
                    cmp(a, cb, env, s)
                    cb$putcode(PUSHARG.OP)
                }
                else if (is.null(a)) 
                    cb$putcode(PUSHNULLARG.OP)
                else if (is.logical(a) && length(a) == 1 && ! is.na(a)) {
                    if (a) cb$putcode(PUSHTRUEARG.OP)
                    else cb$putcode(PUSHFALSEARG.OP)
                }
                else {
                    ci <- cb$putconst(a);
                    cb$putcode(PUSHCONSTARG.OP, ci)
                }
            }
            if (! is.null(n) && n != "") {
                ci <- cb$putconst(as.name(n))
                cb$putcode(SETTAG.OP, ci)
            }
        }
    }
}

cmpBuiltin <- function(e, cb, env, state, internal = FALSE) {
    fun <- e[[1]]
    args <- e[-1]
    names <- names(args)
    if (dots.or.missing(args)) return(FALSE);
    if (typeof(fun) == "character")
        fun <- as.name(fun)
    ci <- cb$putconst(fun)
    if (internal) cb$putcode(GETINTLBUILTIN.OP, ci)
    else cb$putcode(GETBUILTIN.OP, ci)
    cmpBuiltinArgs(args, names, cb, env, state, FALSE)
    ci <- cb$putconst(e)
    cb$putcode(CALLBUILTIN.OP, ci)
    if (state$tailcall) cb$putcode(RETURN.OP)
    TRUE
}

primOps$"%%" <- cmpBuiltin
primOps$"%*%" <- cmpBuiltin
primOps$"%/%" <- cmpBuiltin
primOps$".Alias" <- cmpBuiltin
primOps$".C" <- cmpBuiltin
primOps$".Call" <- cmpBuiltin
primOps$".External" <- cmpBuiltin
primOps$".Fortran" <- cmpBuiltin
primOps$".Primitive" <- cmpBuiltin
primOps$":" <- cmpBuiltin
primOps$"Arg" <- cmpBuiltin
primOps$"Conj" <- cmpBuiltin
primOps$"Im" <- cmpBuiltin
primOps$"Mod" <- cmpBuiltin
primOps$"Re" <- cmpBuiltin
primOps$"abs" <- cmpBuiltin
primOps$"acos" <- cmpBuiltin
primOps$"acosh" <- cmpBuiltin
primOps$"as.call" <- cmpBuiltin
primOps$"asin" <- cmpBuiltin
primOps$"asinh" <- cmpBuiltin
primOps$"atan" <- cmpBuiltin
primOps$"atanh" <- cmpBuiltin
primOps$"attr" <- cmpBuiltin
primOps$"attributes" <- cmpBuiltin
primOps$"attributes<-" <- cmpBuiltin
primOps$"ceiling" <- cmpBuiltin
primOps$"class" <- cmpBuiltin
primOps$"class<-" <- cmpBuiltin
primOps$"cos" <- cmpBuiltin
primOps$"cosh" <- cmpBuiltin
primOps$"cummax" <- cmpBuiltin
primOps$"cummin" <- cmpBuiltin
primOps$"cumprod" <- cmpBuiltin
primOps$"cumsum" <- cmpBuiltin
primOps$"debug" <- cmpBuiltin
primOps$"environment<-" <- cmpBuiltin
primOps$"floor" <- cmpBuiltin
primOps$"gc.time" <- cmpBuiltin
primOps$"globalenv" <- cmpBuiltin
primOps$"invisible" <- cmpBuiltin
primOps$"is.array" <- cmpBuiltin
primOps$"is.atomic" <- cmpBuiltin
primOps$"is.call" <- cmpBuiltin
primOps$"is.character" <- cmpBuiltin
primOps$"is.complex" <- cmpBuiltin
primOps$"is.double" <- cmpBuiltin
primOps$"is.environment" <- cmpBuiltin
primOps$"is.expression" <- cmpBuiltin
primOps$"is.finite" <- cmpBuiltin
primOps$"is.function" <- cmpBuiltin
primOps$"is.infinite" <- cmpBuiltin
primOps$"is.integer" <- cmpBuiltin
primOps$"is.language" <- cmpBuiltin
primOps$"is.list" <- cmpBuiltin
primOps$"is.loaded" <- cmpBuiltin
primOps$"is.logical" <- cmpBuiltin
primOps$"is.matrix" <- cmpBuiltin
primOps$"is.na" <- cmpBuiltin
primOps$"is.name" <- cmpBuiltin
primOps$"is.nan" <- cmpBuiltin
primOps$"is.null" <- cmpBuiltin
primOps$"is.numeric" <- cmpBuiltin
primOps$"is.object" <- cmpBuiltin
primOps$"is.pairlist" <- cmpBuiltin
primOps$"is.real" <- cmpBuiltin
primOps$"is.recursive" <- cmpBuiltin
primOps$"is.single" <- cmpBuiltin
primOps$"is.symbol" <- cmpBuiltin
primOps$"length" <- cmpBuiltin
primOps$"length<-" <- cmpBuiltin
primOps$"list" <- cmpBuiltin
primOps$"pos.to.env" <- cmpBuiltin
primOps$"proc.time" <- cmpBuiltin
primOps$"sign" <- cmpBuiltin
primOps$"sin" <- cmpBuiltin
primOps$"sinh" <- cmpBuiltin
primOps$"symbol.C" <- cmpBuiltin
primOps$"symbol.For" <- cmpBuiltin
primOps$"tan" <- cmpBuiltin
primOps$"tanh" <- cmpBuiltin
#primOps$"trace" <- cmpBuiltin
primOps$"trunc" <- cmpBuiltin
primOps$"unclass" <- cmpBuiltin
primOps$"undebug" <- cmpBuiltin
#primOps$"untrace" <- cmpBuiltin

## **** these were changed relatively recently from special to builtin (2.6?)
primOps$"as.character" <- cmpBuiltin
primOps$"attr<-" <- cmpBuiltin
primOps$"dim" <- cmpBuiltin
primOps$"dim<-" <- cmpBuiltin
primOps$"dimnames" <- cmpBuiltin
primOps$"dimnames<-" <- cmpBuiltin

## **** more recent builtin conversions
primOps$"digamma" <- cmpBuiltin
primOps$"gamma" <- cmpBuiltin
primOps$"lgamma" <- cmpBuiltin
primOps$"log1p" <- cmpBuiltin
primOps$"trigamma" <- cmpBuiltin
primOps$"interactive" <- cmpBuiltin
primOps$"browser" <- cmpBuiltin
primOps$"nargs" <- cmpBuiltin
cmpSpecial <- function(e, cb, env, state) {
    fun <- e[[1]]
    if (typeof(fun) == "character")
        fun <- as.name(fun)
    ci <- cb$putconst(e)
    cb$putcode(CALLSPECIAL.OP, ci)
    if (state$tailcall) cb$putcode(RETURN.OP)
    TRUE
}

primOps$"$" <- cmpSpecial
primOps$"$<-" <- cmpSpecial
primOps$".Internal" <- cmpSpecial
primOps$"<<-" <- cmpSpecial
primOps$"UseMethod" <- cmpSpecial
primOps$"[" <- cmpSpecial
primOps$"[<-" <- cmpSpecial
primOps$"[[" <- cmpSpecial
primOps$"[[<-" <- cmpSpecial
primOps$"c" <- cmpSpecial
primOps$"call" <- cmpSpecial
primOps$"expression" <- cmpSpecial
primOps$"missing" <- cmpSpecial
primOps$"on.exit" <- cmpSpecial
primOps$"return" <- cmpSpecial
primOps$"substitute" <- cmpSpecial
primOps$"~" <- cmpSpecial
primOps$"log" <- cmpSpecial
primOps$"<-" <- function(e, cb, env, state) {
    rhs <- e[[2]]
    if (is.character(rhs)) rhs <- as.name(rhs)
    if (is.name(rhs)) {
        s<-make.argState(state)
        cmp(e[[3]], cb, env, s)
        ci <- cb$putconst(rhs)
        cb$putcode(SETVAR.OP, ci)
        if (state$tailcall) {
            cb$putcode(INVISIBLE.OP)
            cb$putcode(RETURN.OP)
        }
        TRUE
    }
    else if (typeof(rhs) == "language" && is.symbol(rhs[[2]])) {
        nvalue <- place <- rhs
        value <- e[[3]]
        symbol <- place[[2]]
        vsym <- as.name("*ctmp*");
        nvalue[[1]] <- as.name(paste(place[[1]],"<-", sep=""))
        #**** place$value <- value only works if value is not NULL
        nvalue$value <- NULL # clear existing value component
        nvalue$value <- vsym;
        s <- make.argState(state)
        cmp(value, cb, env, s)
        cvi <- cb$putconst(vsym)
        csi <- cb$putconst(symbol)
        cb$putcode(STARTASSIGN.OP, csi, cvi);
        cmp(nvalue, cb, env, s)
        cb$putcode(ENDASSIGN.OP, csi, cvi)
        if (state$tailcall) {
            cb$putcode(INVISIBLE.OP)
            cb$putcode(RETURN.OP)
        }
        TRUE;
    }
    else cmpSpecial(e, cb, env, state)
}
primOps$"=" <- primOps$"<-"
primOps$"&&" <- function(e, cb, env, state) {
    ee <- substitute(if (a) { if (b) TRUE else FALSE } else FALSE,
                     list(a=e[[2]], b=e[[3]]))
    cmp(ee, cb, env, state)
    TRUE
}

primOps$"||" <- function(e, cb, env, state) {
    ee <- substitute(if (a) TRUE else if (b) TRUE else FALSE,
                     list(a=e[[2]], b=e[[3]]))
    cmp(ee, cb, env, state)
    TRUE
}
primOps$"==" <- function(e, cb, env, state) {
   cmpPrim2(e, cb, env, EQ.OP, state);
}

primOps$"!=" <- function(e, cb, env, state) {
   cmpPrim2(e, cb, env, NE.OP, state);
}

primOps$"<" <- function(e, cb, env, state) {
   cmpPrim2(e, cb, env, LT.OP, state);
}

primOps$"<=" <- function(e, cb, env, state) {
   cmpPrim2(e, cb, env, LE.OP, state);
}

primOps$">=" <- function(e, cb, env, state) {
   cmpPrim2(e, cb, env, GE.OP, state);
}

primOps$">" <- function(e, cb, env, state) {
   cmpPrim2(e, cb, env, GT.OP, state);
}

primOps$"&" <- function(e, cb, env, state) {
   cmpPrim2(e, cb, env, AND.OP, state);
}

primOps$"|" <- function(e, cb, env, state) {
   cmpPrim2(e, cb, env, OR.OP, state);
}

primOps$"!" <- function(e, cb, env, state) {
   cmpPrim1(e, cb, env, NOT.OP, state);
}
cmpDispatch <- function(start.op, end.op, e, cb, env, state) {
    if (any.dots(e))
        cmpSpecial(e, cb, env, state)
    else {
        s<-make.argState(state)
        ne <- length(e)
        oe <- if (ne > 1) e[[2]] else NULL
        cmp(oe, cb, env, s)
        ci <- cb$putconst(e)
        end.label <- cb$makelabel()
        cb$putcode(start.op, ci, end.label)
        if (ne > 2)
            cmpBuiltinArgs(e[-(1:2)], names(e)[-(1:2)], cb, env, state, TRUE)
        cb$putcode(end.op)
        cb$putlabel(end.label)
        if (state$tailcall) cb$putcode(RETURN.OP)
        TRUE
    }
}

primOps$"[" <- function(e, cb, env, state)
    cmpDispatch(STARTSUBSET.OP, DFLTSUBSET.OP, e, cb, env, state)

primOps$"[<-" <- function(e, cb, env, state)
    cmpDispatch(STARTSUBASSIGN.OP, DFLTSUBASSIGN.OP, e, cb, env, state)

primOps$"c" <- function(e, cb, env, state)
    cmpDispatch(STARTC.OP, DFLTC.OP, e, cb, env, state)

primOps$"[[" <- function(e, cb, env, state)
    cmpDispatch(STARTSUBSET2.OP, DFLTSUBSET2.OP, e, cb, env, state)

primOps$"[[<-" <- function(e, cb, env, state)
    cmpDispatch(STARTSUBASSIGN2.OP, DFLTSUBASSIGN2.OP, e, cb, env, state)
primOps$"$"<- function(e, cb, env, state) {
    if (any.dots(e) || length(e) != 3)
        cmpSpecial(e, cb, env, state)
    else {
        sym <- if (is.character(e[[3]])) as.name(e[[3]]) else e[[3]]
        if (is.name(sym)) {
            s<-make.argState(state)
            cmp(e[[2]], cb, env, s)
            ci <- cb$putconst(e)
            csi <- cb$putconst(sym)
            cb$putcode(DOLLAR.OP, ci, csi)
            if (state$tailcall) cb$putcode(RETURN.OP)
            TRUE
        }
        else cmpSpecial(e, cb, env, state)
    }
}

primOps$"$<-"<- function(e, cb, env, state) {
    if (any.dots(e) || length(e) != 4)
        cmpSpecial(e, cb, env, state)
    else {
        sym <- if (is.character(e[[3]])) as.name(e[[3]]) else e[[3]]
        if (is.name(sym)) {
            s<-make.argState(state)
            cmp(e[[4]], cb, env, s)
            cmp(e[[2]], cb, env, s)
            ci <- cb$putconst(e)
            csi <- cb$putconst(sym)
            cb$putcode(DOLLARGETS.OP, ci, csi)
            if (state$tailcall) cb$putcode(RETURN.OP)
            TRUE
        }
        else cmpSpecial(e, cb, env, state)
    }
}
cmpIs <- function(op, e, cb, env, state) {
    if (any.dots(e) || length(e) != 2)
        cmpBuiltin(e, cb, env, state)
    else {
        s<-make.argState(state)
        cmp(e[[2]], cb, env, s)
        cb$putcode(op)
        if (state$tailcall) cb$putcode(RETURN.OP)
        TRUE
    }
}

primOps$"is.character" <- function(e, cb, env, state)
    cmpIs(ISCHARACTER.OP, e, cb, env, state)
primOps$"is.complex" <- function(e, cb, env, state)
    cmpIs(ISCOMPLEX.OP, e, cb, env, state)
primOps$"is.double" <- function(e, cb, env, state)
    cmpIs(ISDOUBLE.OP, e, cb, env, state)
primOps$"is.integer" <- function(e, cb, env, state)
    cmpIs(ISINTEGER.OP, e, cb, env, state)
primOps$"is.logical" <- function(e, cb, env, state)
    cmpIs(ISLOGICAL.OP, e, cb, env, state)
primOps$"is.name" <- function(e, cb, env, state)
     cmpIs(ISSYMBOL.OP, e, cb, env, state)
primOps$"is.null" <- function(e, cb, env, state)
    cmpIs(ISNULL.OP, e, cb, env, state)
primOps$"is.numeric" <- function(e, cb, env, state)
    cmpIs(ISNUMERIC.OP, e, cb, env, state)
primOps$"is.object" <- function(e, cb, env, state)
    cmpIs(ISOBJECT.OP, e, cb, env, state)
primOps$"is.real" <- function(e, cb, env, state)
    cmpIs(ISDOUBLE.OP, e, cb, env, state)
primOps$"is.symbol" <- function(e, cb, env, state)
    cmpIs(ISSYMBOL.OP, e, cb, env, state)
primOps$".Internal" <- function(e, cb, env, state) {
    ee <- e[[2]]
    sym <- ee[[1]]
    if (.Internal(is.builtin.internal(sym))) {
        # cat(paste("compiling call to .Internal(", sym, ")\n", sep=""))
        cmpBuiltin(ee, cb, env, state, internal=TRUE)
    }
    else cmpSpecial(e, cb, env, state);
}
.Topenv<-.GlobalEnv
cmpSimpleInternal <- function(e, cb, env, state) {
    if (any.dots(e))
        FALSE
    else {
        name <- as.character(e[[1]])
        # notifyInline(name);
        def <- get(name, env=.Topenv)
        if (! checkCall(def, e, NULL)) return(FALSE)
        cenv <- c(as.list(match.call(def, e, F))[-1], formals(def))
        lcall <- c(e[[1]],cenv[names(formals(def))])
        names(lcall)<-NULL
        call <- as.call(lcall)
        cmp(substitute(.Internal(call), list(call=call)), cb, env, state)
        TRUE
    }
}

notifyInline <- function(fun) cat(paste("inlining call to", fun, "\n"))

# primOps$log<-function(e, cb, env, state) {
#     if (any.dots(e))
#         FALSE
#     else {
#         n <- length(e)
#         def <- get("log",env=.BaseEnv)
#         if (! checkCall(def, e, NULL)) return(FALSE)
#         call <- match.call(def, e)
#         if (n == 2 || n == 3) {
#            # notifyInline("log");
#            cmp(substitute(.Internal(call), list(call=call)), cb, env, state)
#            TRUE
#         }
#         else FALSE
#     }
# }

cmpMaxLike <- function(e, cb, env, state) {
    if (any.dots(e))
        FALSE
    else {
        name <- as.character(e[[1]])
        def <- get(name, env=.Topenv)
        if (! checkCall(def, e, NULL)) return(FALSE)
        call <- match.call(def, e)
        if (is.null(call$na.rm)) {
            lcall <- as.list(call)
            lcall$na.rm<-FALSE
            call <- as.call(lcall)
        }
        # notifyInline(name);
        cmp(substitute(.Internal(call), list(call=call)), cb, env, state)
        TRUE
    }
}

primOps$max <- cmpMaxLike    
primOps$min <- cmpMaxLike    
primOps$all <- cmpMaxLike    
primOps$any <- cmpMaxLike    
primOps$sum <- cmpMaxLike    
primOps$prod <- cmpMaxLike    
primOps$inherits <- cmpSimpleInternal

#******* need to avoid using body since that gets replaced!!

safeInternals <- c("atan2", "besselJ", "besselY", "beta", "chartr",
  "choose", "dbinom", "dcauchy","df", "dgamma", "dgeom", "dhyper",
  "digamma", "dlnorm", "dlogis", "dnbinom", "dnorm", "dpois",
  "drop", "dsignrank", "dt", "dunif", "dweibull", "dwilcox",
  "fft", "gamma", "gammaCody", "inherits", "is.vector", "lbeta", 
  "lchoose", "lgamma", "log1p", "mvfft", "nchar", "nextn", "pbinom",
  "pcauchy", "pentagamma", "pgamma", "pgeom", "phyper", "plnorm",
  "plogis", "pnbinom", "pnorm", "polyroot", "ppois", "psignrank",
  "ptukey", "punif", "pweibull", "pwilcox", "qbeta",  "qbinom",
  "qcauchy", "qf", "qgamma", "qgeom", "qhyper", "qlnorm", "qlogis",
  "qnbinom", "qnorm", "qpois", "qsignrank", "qt", "qtukey", "qunif",
  "qweibull", "qwilcox", "rbeta", "rbinom", "rcauchy", "rf", "rgamma", 
  "rgb", "rgeom", "rhyper", "rlnorm", "rlogis", "rnbinom", "rnorm", 
  "round", "rpois", "rsignrank", "rt", "runif", "rweibull", "rwilcox", 
  "signif", "tetragamma", "tolower", "toupper", "trigamma", "typeof",
  "vector", "which.max", "which.min")

## updated list -- need programatic check that these are OK
safeInternals <- c("atan2", "besselY", "beta", "choose", 
  "dbinom",  "dcauchy", "dgeom", "dhyper", "dlnorm", "dlogis",
  "dnorm", "dpois", "drop", "dunif", "dweibull", "fft", "inherits",
  "is.vector", "lbeta", "lchoose", "mvfft", "nchar", "nextn", "pbinom",
  "pcauchy", "pgeom", "phyper", "plnorm", "plogis", "pnorm",
  "polyroot", "ppois", "punif", "pweibull", "qbinom", "qcauchy", 
  "qgamma", "qgeom", "qhyper", "qlnorm", "qlogis", "qnorm", "qpois", 
  "qunif", "qweibull", "rbinom", "rcauchy", 
  "rgeom", "rhyper", "rlnorm", "rlogis", "rnorm", "rpois", "rsignrank", 
  "runif", "rweibull", "rwilcox", "typeof", "vector", "which.max", 
  "which.min")

## "ptukey", "qtukey",  are kind of safe but reverse some args.

for (i in safeInternals) primOps[[i]] <- cmpSimpleInternal
primOps$"function" <- function(e, cb, env, state) {
    forms <- e[[2]]
    body <- e[[3]]
    nenv <- funEnv(forms, body, env)
    cbody <- compile(body, nenv)
    ci <- cb$putconst(list(forms, cbody))
    cb$putcode(MAKECLOSURE.OP, ci)
    if (state$tailcall) cb$putcode(RETURN.OP)
    TRUE
}

is.ddsym <- function(name) {
    (is.symbol(name) || is.character(name)) &&
    length(grep("^\\.\\.[0-9]*$", as.character(name))) != 0
}

punt <- function(e) stop(paste("can't compile this yet:", as.expression(e)))

##**** need to handle ... and ..n arguments specially
##**** separate call opcode for calls with named args?
## **** for (a in e[[-1]]) ... goes into infinite loop

cmpCallArgs <- function(e, cb, env, state) {
    args <- e[-1]
    argc <- length(args)
     if (argc > 0) {
        names <- names(args)
        do.tag <- function(n) {
            if (! is.null(n) && n != "") {
                ci <- cb$putconst(as.name(n))
                cb$putcode(SETTAG.OP, ci)
            }
        }
        for (i in 1:length(args)) {
            a <- args[[i]]
            n <- names[[i]]
            if (missing(a)) { #**** better test for missing??
                cb$putcode(DOMISSING.OP)
                do.tag(n)
            }
            else if (is.symbol(a) && a == "...") {
                if (! findVar("...", env))
                    cb$warn("... may be used in an incorrect context");
                cb$putcode(DODOTS.OP)
            }
            else if (is.symbol(a) || typeof(a) == "language") {
                ci <- cb$putconst(compile(a, env, promise=TRUE))
                cb$putcode(MAKEPROM.OP, ci)
                do.tag(n)
             }
             else { 
                ci <- cb$putconst(a)
                cb$putcode(PUSHCONSTARG.OP, ci)
                do.tag(n)
             }
        }
    }
}

checkCall <- function(def, call, signal = warning) {
    if (any.dots(call))
        NA
    else {
        old <-options()$show.error.messages
        if (is.null(old)) old <- TRUE
        options(show.error.messages=FALSE)
        msg <- try({match.call(def, call); NULL})
        options(show.error.messages=old)
        if (! is.null(msg)) {
            msg<-sub("\n$","",sub("^E.*: ","",msg))
            emsg<-paste("possible error in ", deparse(call),": ", msg, sep="")
            if (! is.null(signal)) signal(emsg)
            FALSE
        }
        else TRUE
    }
}


cmpCall <- function(e, cb, env, state) {
    fun <- e[[1]]
    if (typeof(fun) == "character")
        fun <- as.name(fun)
    if (typeof(fun) == "symbol") {
        ci <- cb$putconst(fun)
        if (findVar(fun, env)) {
            cb$add.locfun(fun)
            cb$putcode(GETFUN.OP, ci)
        }
        else {
            name <- as.character(fun)
            fenv <- if (is.null(env[[2]])) .GlobalEnv else env[[2]]
            if (exists(name, env=fenv)) {
                def <- get(name, env=fenv)
                if (typeof(def) == "closure")
                    checkCall(def, e, cb$warn)
            }
            else cb$add.undef(fun)
            if (isSysVar(name, env) && exists(name,env=.BaseEnv))
                cb$putcode(GETSYMFUN.OP, ci)
            else if (isNamespace(env[[2]]))
                cb$putcode(GETFUN.OP, ci)
            else cb$putcode(GETGLOBFUN.OP, ci)
        }
    }
    else {
        if (! is.null(constantFold(fun, env)))
            cb$warn(paste("possible call to non-function:", deparse(e)))
        s <- make.argState(state)
        cmp(fun, cb, env, s)
        cb$putcode(CHECKFUN.OP)
    }
    cmpCallArgs(e, cb, env, state)
    ci <- cb$putconst(e)
    cb$putcode(CALL.OP, ci)
    if (state$tailcall) cb$putcode(RETURN.OP)
}

cmpSym <- function(e, cb, env, state) {
    if (e == "...") {
        # warning on compile and runtime error??
        cb$warn("... may be used in an incorrect context");
        cb$putcode(DOTSERR.OP)
    }
    else if (is.ddsym(e)) {
        if (! findVar("...", env))
            cb$warn(paste(e, "may be used in an incorrect context"))
        ci <- cb$putconst(e)
        cb$putcode(DDVAL.OP, ci)
        if (state$tailcall) cb$putcode(RETURN.OP)
    }
    else {
        if (! findVar(e, env) && e != "*tmp*" && e != "*ctmp*")
            cb$add.globvar(e)
        ci <- cb$putconst(e)
        cb$putcode(GETVAR.OP, ci)
        if (state$tailcall) cb$putcode(RETURN.OP)
    }
}

primNames<-names(primOps)

cmp <- function(e, cb, env, state) {
    ce <- constantFold(e, env)
        if (! is.null(ce)) e <- ce
    if (typeof(e) == "language") {
        fun <- e[[1]]
        if (typeof(fun) == "symbol") {
            if (isSysVar(fun, env)) {
                pi <- matchName(fun, primNames)
                p <- if (pi) primOps[[pi]]
                if (is.null(p) || ! p(e, cb, env, state))
                    cmpCall(e, cb, env, state)
            }
            else cmpCall(e, cb, env, state)
        }
        else cmpCall(e, cb, env, state)
    }
    else if (typeof(e) == "symbol")
        cmpSym(e, cb, env, state)
    else if (is.null(e)) {
        cb$putcode(LDNULL.OP)
        if (state$tailcall) cb$putcode(RETURN.OP)
    }
    else if (is.logical(e) && length(e) == 1 && ! is.na(e)) {
        if (e) cb$putcode(LDTRUE.OP)
        else cb$putcode(LDFALSE.OP)
        if (state$tailcall) cb$putcode(RETURN.OP)
    }
    else {
        ci <- cb$putconst(e)
        cb$putcode(LDCONST.OP, ci)
        if (state$tailcall) cb$putcode(RETURN.OP)
    }
}

compile <- function(e, env=NULL, promise = FALSE) {
    cb <- make.codeBuf()
    #if (promise) cb$putconst(e)
    cb$putconst(e) #**** put expression in as first constant.
    cmp(e, cb, env, list(tailcall=TRUE))
    globvars <- cb$globvars()
    if (length(globvars) != 0)
        cb$warn(paste("global variables used:", paste(globvars,collapse=",")))
    locfuns <- cb$locfuns()
    if (length(locfuns) != 0)
        cb$warn(paste("local functions used:", paste(locfuns,collapse=",")))
    undefs <- cb$undefs()
    if (length(undefs) != 0)
        cb$warn(paste("undefined functions used:", paste(undefs,collapse=",")))
    .Internal(mkCode(cb$code(),cb$const()))
}
    
lexvars <- function(env) {
    top <- topenv(env)
    lex <- NULL
    while (! identical(env, top)) {
        lex <- unique(c(ls(env, all = TRUE), lex))
        env <- parent.env(env)
    }
    lex
}

cmpfun <- function(f, env = NULL) {
    type <- typeof(f)
    if (type == "closure") {
        env <- unique(c(env, lexvars(environment(f))))
        env <- list(env, topenv(environment(f)))
        nenv <- funEnv(formals(f), body(f), env)
        b <- compile(body(f), nenv)
        .Internal(bcClose(formals(f), b, environment(f)))
    }
    else if (typeof(f) == "builtin" || type == "special")
        f
    else stop("can't compile a non-function")
}
cmpframe <- function(inpos, file) {
    expr.needed <- 1000
    expr.old <- options()$expressions
    if (expr.old < expr.needed)
       options(expressions = expr.needed)
    on.exit(options(expressions = expr.old))

    attach(NULL, name="<compiled>")
    inpos <- inpos + 1
    outpos <- 2
    on.exit(detach(pos=outpos), add=TRUE)

    for (f in ls(pos=inpos,all=TRUE)) {
        def <- get(f, pos=inpos)
        if (typeof(def) == "closure") {
                cat(paste("compiling", f, "\n"))
                fc <- cmpfun(def)
                assign(f, fc, pos=outpos)
        }
    }
    cat(paste("saving to file \"", file, "\" ... ", sep=""))
    save(list=ls(pos=outpos,all=T), file=file)
    cat("done\n")
}

cmplib <- function(package, file) {
    package <- as.character(substitute(package))
    pkgname <- paste("package", package, sep = ":")
    pos <- match(pkgname, search());
    if (missing(file))
        file <- paste(package,".Rc",sep="")
    if (is.na(pos)) {
        library(package, char=TRUE)
        pos <- match(pkgname, search());
        on.exit(detach(pos=match(pkgname, search())))
    }
    cmpframe(pos, file)
}

cmpfile <- function(infile, outfile, ascii = FALSE, env = .GlobalEnv) {
    if (! is.environment(env) || ! identical(env, topenv(env)))
        stop("env must be a top level environment")
    if (missing(outfile)) {
        basename <- sub("\\.[a-zA-Z0-9]$", "", infile)
        outfile <- paste(basename, ".Rc", sep="")
    }
    if (infile == outfile)
        stop("input and output file names are the same")
    forms <- parse(infile)
    nforms <- length(forms)
    if (nforms > 0) {
    expr.needed <- 1000
    expr.old <- options()$expressions
        if (expr.old < expr.needed) {
            options(expressions = expr.needed)
            on.exit(options(expressions = expr.old))
        }
        cforms <- vector("list", nforms)
        for (i in 1:nforms) {
            e <- forms[[i]]
            if (typeof(e) == "language" && e[[1]] == "<-" &&
                typeof(e[[3]]) == "language" && e[[3]][[1]] == "function")
                cat(paste("compiling function \"", e[[2]], "\"\n", sep=""))
            else
                cat(paste("compiling expression", deparse(e, 20)[1], "...\n"))
            cforms[[i]] <- compile(e, list(character(0), env))
        }
        cat(paste("saving to file \"", outfile, "\" ... ", sep=""))
        .Internal(save.to.file(cforms, outfile, ascii))
        cat("done\n")
    }
    else warning(" empty input file; no output written");
    invisible(NULL)
}

loadcmp <- function (file, envir = .GlobalEnv, chdir = FALSE) 
{
    if (!(is.character(file) && file.exists(file))) 
        stop(paste("file `", file, "' does not exist", sep = ""))
    exprs <- .Internal(load.from.file(file))
    if (length(exprs) == 0) 
        return(invisible())
    if (chdir && (path <- dirname(file)) != ".") {
        owd <- getwd()
        on.exit(setwd(owd), add = TRUE)
        setwd(path)
    }
    for (i in exprs) {
        yy <- eval(i, envir)
    }
    invisible()
}
bcDecode <- function(code) {
    n <- length(code)
    ncode <- vector("list", n)
    ncode[[1]] <- code[1] # version number
    i <- 2
    while (i <= n) {
        name<-Opcodes.names[code[i]+1]
        argc<-Opcodes.argc[[code[i]+1]]
        ncode[[i]] <- as.name(name)
        i<-i+1
        if (argc > 0)
            for (j in 1:argc) {
                ncode[[i]]<-code[i]
                i<-i+1
            }
    }
    ncode
}
    
disassemble <- function(code) {
    .CodeSym <- as.name(".Code")
    disasm.const<-function(x)
        if (typeof(x)=="list" && length(x) > 0 && identical(x[[1]], .CodeSym))
            disasm(x) else x
    disasm <-function(code) {
        code[[2]]<-bcDecode(code[[2]])
        code[[3]]<-lapply(code[[3]], disasm.const)
        code
    }
    if (typeof(code)=="closure") {
        code <- .Internal(bodyCode(code))
        if (typeof(code) != "bytecode")
            stop("functions is not compiled")
    }
    dput(disasm(.Internal(disassemble(code))))
}
